#!/usr/local/bin/perl -w

# read gismu_formatted.csv, build up the two databases we need, write the
# two description files for use by prc.pl

# each line of the input file looks like:
#  gapru,x1 is,above/upwards,from x2,in frame of reference x3,,,2d,202
# the comma-separated fields are:
#  gismu, x1-clause, description, x2-clause, x3-clause, x4-clause, x5-clause,
#  section, count
# actually, there may be commas inside ""-bracketed fields, so we can't just
# use split().

# for each clause, replace /x\d/ with a marker ('%' comes to mind). Then build
# a hash of clauses {clause}. Create a record for the gismu: {x1, x2, x3, x4,
# x5, section (two letters), count, description}. The x1,x2,etc are hashrefs
# to the clause hash.

# when all the gismu are done, sort the clause hash by clause and assign each
# clause a number (0..), storing the number in $clause->{'number'}.
# sort the gismu alphabetically (they will be used in a binary search)

# then emit the clause description file (GismuPlacesDB.txt). Each record looks
# like this:
# struct GismuPlaceCluster {
#   char offsets[4];
#   char firstplace; // 4 null-terminated strings
# }

# then emit the gismu description file (GismuDB.txt). Each record is as
# follows:
#  struct GismuRecord {
#      char gismu[5];
#      uint16 places[5]; // 0 for none, 1..1300ish otherwise
#      uint16 count;     // 0 for very infrequent
#      char numrafsi;
#      char numrefs;
#      char rafsi[3][numrafsi]; // bit7 of last letter set if apostrophe needed
#      uint16 refs[numrefs]; // gismu numbers
#      char description[]; // null-terminated
#  } # total size 15+desc+3*rafsi+2*refs per gismu
# struct GismuCluster {
#      char offsets[4];
#      struct GismuRecord gismu[4];
# } # 4 per 4 gismu

# clustering: GismuDB: 40s, GismuPlacesDB: 37s (was 8min each)
# new: 1m07, 1m11

use Lingua::Lojban qw(gismu rafsilist rafsi2gismu);

open(CSV, "<gismu_formatted.csv") or die;

while(<CSV>) {
    s/[\n\r]+$//;
    if (0) {
	my(@bits);
	my $copy = $_;
	while(length($_)) {
	    if (0 and /^\"/) {
		# looks like /"foo",/, so absorb up till the next /",/
		s/^\"(.*)\"($|,)// or die "weird";
		push(@bits, $1);
	    } else {
		/^([^,]*)($|,)/ or die "weird";
		push(@bits, $1);
		s/^([^,]*)($|,)// or die "weird";
	    }
	}
    }
    my($gismu,$x1,$desc,$x2,$x3,$x4,$x5,$section,$count) = split(/@/,$_);
    unless (defined($count)) {
	print "bad line\n";
	#print "$copy\n";
	#print "(",join('|',@bits),")\n";
	print "$_\n";
	die;
    }
    
    my($gis) = gismu($gismu) or die;

    my($c,$g);
    $g = { 'desc' => $desc,
	   'section' => $section,
	   'count' => $count,
	   'rafsi1' => $gis->{'rafsi1'},
	   'rafsi2' => $gis->{'rafsi2'},
	   'rafsi3' => $gis->{'rafsi3'},
	   'refs' => $gis->{'refs'},
	 };

    # the null clause will sort to position 0
    $x1 =~ s/\bx1\b/%/g;
    $clauses{$x1} = { 'clause' => $x1 } unless $clauses{$x1};
    $g->{'x1'} = $clauses{$x1};

    $x2 =~ s/\bx2\b/%/g;
    $clauses{$x2} = { 'clause' => $x2 } unless $clauses{$x2};
    $g->{'x2'} = $clauses{$x2};

    $x3 =~ s/\bx3\b/%/g;
    $clauses{$x3} = { 'clause' => $x3 } unless $clauses{$x3};
    $g->{'x3'} = $clauses{$x3};

    $x4 =~ s/\bx4\b/%/g;
    $clauses{$x4} = { 'clause' => $x4 } unless $clauses{$x4};
    $g->{'x4'} = $clauses{$x4};

    $x5 =~ s/\bx5\b/%/g;
    $clauses{$x5} = { 'clause' => $x5 } unless $clauses{$x5};
    $g->{'x5'} = $clauses{$x5};

    $gismu{$gismu} = $g;
}

close(CSV);

sub rafsicode {
    my($rafsi) = @_;
    if (length($rafsi) == 3) {
	return $rafsi;
    } else {
	return substr($rafsi, 0, 2) . chr(0x80 | ord(substr($rafsi,3,1)));
    }
}

# sort clauses
@clauses = sort {$clauses{$a}->{'clause'} cmp $clauses{$b}->{'clause'} }
  keys(%clauses);
foreach $i (0 .. $#clauses) {
    $clauses{$clauses[$i]}->{'number'} = $i;
}

# sort gismu
@gismu = sort(keys(%gismu));
foreach my $i (0 .. $#gismu) {
    $gismulist{$gismu[$i]} = $i;
}

# now emit clause description
# ...
if (1) {
    open(C,">c.out") or die;
    print C <<'EOF';
name GismuPlacesDB
record-db
read-only
type GisP
creator Gsmu
version 0
modify-count 1
id-seed 0
create-time 98-11-11 14:07:00
modify-time 98-11-11 14:07:00
backup-time 98-11-11 14:07:00

# app-info stub

app-info
0000: 00 60 55 6e 66 69 6c 65 64 00 00 00 00 00 00 00
0010: 00 00 42 75 73 69 6e 65 73 73 00 00 00 00 00 00      ..Business......
0020: 00 00 50 65 72 73 6f 6e 61 6c 00 00 00 00 00 00      ..Personal......
0030: 00 00 50 69 6c 6f 74 20 50 72 6f 6a 65 63 74 73      ..Pilot Projects
0040: 00 00 50 65 72 6c 20 50 72 6f 6a 65 63 74 73 00      ..Perl Projects.
0050: 00 00 41 64 64 72 65 73 73 20 55 70 64 61 74 65      ..Address Update
0060: 73 00 42 69 6c 6c 73 20 74 6f 20 53 70 6c 69 74      s.Bills to Split
0070: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0080: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0090: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00a0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00b0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00c0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00d0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00e0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00f0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0100: 00 00 00 01 02 10 11 13 14 00 00 00 00 00 00 00      ................
0110: 00 00 14 00 00 00 ff ff 01 00                        ..........


EOF

    print C "# ",scalar(@clauses)," records\n";
    for (my $count = 0; $count < $#clauses; $count += 4) {
	my(@r,$s);
	print C "# $count .. ",$count+3,"\n";

	my $ptr = 0;
	for (my $subcount = 0;
	     $subcount < 4 and $count+$subcount < $#clauses;
	     $subcount++)
	{
	    my $c = $clauses{$clauses[$count+$subcount]}->{'clause'};
	    $r[$subcount] = $c . "\0";
	    die "big ptr" if $ptr > 255;
	    $s .= pack("C",$ptr);
	    $ptr += length($r[$subcount]);
	    print C "# ",$count+$subcount,": ";
	    print C $clauses{$clauses[$count+$subcount]}->{'clause'},"\n";
	}
	printf C "record cat=\@ id=0x00%04X archive\n", $count/4;
	print C "0: ",uc(unpack("H*",$s)),"\n";
	foreach my $subcount (0 .. 3) {
	    print C $subcount+1,": ",uc(unpack("H*",$r[$subcount])),"\n"
	      if $r[$subcount];
	}
	print C "\n";
    }
    close(C);
}

# now emit gismu description
# ...
if (1) {
    open(G,">g.out") or die;
    print G <<'EOF';
name GismuDB
record-db
read-only
type GisD
creator Gsmu
version 0
modify-count 1
id-seed 0
create-time 98-11-11 14:07:00
modify-time 98-11-11 14:07:00
backup-time 98-11-11 14:07:00

# app-info stub

app-info
0000: 00 60 55 6e 66 69 6c 65 64 00 00 00 00 00 00 00      ..Unfiled.......
0010: 00 00 42 75 73 69 6e 65 73 73 00 00 00 00 00 00      ..Business......
0020: 00 00 50 65 72 73 6f 6e 61 6c 00 00 00 00 00 00      ..Personal......
0030: 00 00 50 69 6c 6f 74 20 50 72 6f 6a 65 63 74 73      ..Pilot Projects
0040: 00 00 50 65 72 6c 20 50 72 6f 6a 65 63 74 73 00      ..Perl Projects.
0050: 00 00 41 64 64 72 65 73 73 20 55 70 64 61 74 65      ..Address Update
0060: 73 00 42 69 6c 6c 73 20 74 6f 20 53 70 6c 69 74      s.Bills to Split
0070: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0080: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0090: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00a0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00b0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00c0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00d0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00e0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00f0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0100: 00 00 00 01 02 10 11 13 14 00 00 00 00 00 00 00      ................
0110: 00 00 14 00 00 00 ff ff 01 00                        ..........

EOF

  print G "# ",scalar(@gismu)," gismu\n";
    for (my $count = 0; $count < $#gismu; $count += 4) {
	my(@r,$s);
	for (my $subcount = 0; 
	     $subcount < 4 and $count+$subcount < $#gismu;
	     $subcount++) 
	{
	    my $gismu = $gismu[$count+$subcount];
#	    print "doing $count+$subcount: $gismu\n";
	    my $g = $gismu{$gismu};
	    my(@p) = map { $g->{$_}->{'number'} } qw(x1 x2 x3 x4 x5);
	    print G "# $gismu: ",join(',',@p),"\n";
	    $r[$subcount] = $gismu . pack("n5",@p) . pack("n",$g->{'count'});
	    my $numrafsi = 0;
	    my $rafsi;
	    foreach my $r (qw(rafsi1 rafsi2 rafsi3)) {
		if ($g->{$r}) {
		    $numrafsi++;
		    $rafsi .= rafsicode($g->{$r});
		}
	    }
	    $r[$subcount] .= pack("C",$numrafsi);
	    $r[$subcount] .= pack("C", 0); # numrefs: do it later
	    $r[$subcount] .= $rafsi if $numrafsi;
	    # split up refs, insert here
	    $r[$subcount] .= $g->{'desc'} . "\0";
	}
	# now make cluster header
	my $ptr = 4;
	foreach my $subcount (0 .. 3) {
	    if ($r[$subcount]) {
		die "big ptr" if $ptr > 255;
		$s .= pack("C",$ptr);
		$ptr += length($r[$subcount]);
	    } else {
		$s .= pack("C",0);
	    }
	}

	print G "# ",$gismu[$count]," ... ",$gismu[$count+3],"\n";
	printf G "record cat=\@ id=0x00%04x archive\n",$count/4;
        print G "0: ",uc(unpack("H*",$s)),"\n";
	foreach my $subcount (0 .. 3) {
	    print G $subcount+1,": ",uc(unpack("H*",$r[$subcount])),"\n"
	      if $r[$subcount];
	}
        print G "\n";
    }
    close(G);
}

# now emit rafsi index: rafsi -> gismu number
#  struct RafsiCluster {
#      struct {
#          unsigned char r1,r2,r3;
#      } rafsi[4];
#      struct {
#          unsigned char high, low;
#      } gismu[4];
#  }

if (1) {
    open(R,">r.out") or die;
    print R <<'EOF';
name RafsiDB
record-db
read-only
type Rfsi
creator Gsmu
version 0
modify-count 1
id-seed 0
create-time 98-11-11 14:07:00
modify-time 98-11-11 14:07:00
backup-time 98-11-11 14:07:00

# app-info stub

app-info
0000: 00 60 55 6e 66 69 6c 65 64 00 00 00 00 00 00 00
0010: 00 00 42 75 73 69 6e 65 73 73 00 00 00 00 00 00      ..Business......
0020: 00 00 50 65 72 73 6f 6e 61 6c 00 00 00 00 00 00      ..Personal......
0030: 00 00 50 69 6c 6f 74 20 50 72 6f 6a 65 63 74 73      ..Pilot Projects
0040: 00 00 50 65 72 6c 20 50 72 6f 6a 65 63 74 73 00      ..Perl Projects.
0050: 00 00 41 64 64 72 65 73 73 20 55 70 64 61 74 65      ..Address Update
0060: 73 00 42 69 6c 6c 73 20 74 6f 20 53 70 6c 69 74      s.Bills to Split
0070: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0080: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0090: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00a0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00b0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00c0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00d0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00e0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
00f0: 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00      ................
0100: 00 00 00 01 02 10 11 13 14 00 00 00 00 00 00 00      ................
0110: 00 00 14 00 00 00 ff ff 01 00                        ..........


EOF
    my(@rafsi) = rafsilist();
    # filter all those that don't have gismu in our database
    @rafsi = grep { 
	defined(rafsi2gismu($_)) and defined($gismulist{rafsi2gismu($_)})
    } @rafsi;
    @rafsi = sort { rafsicode($a) cmp rafsicode($b) } @rafsi;
    print R "# ",scalar(@rafsi)," rafsi\n";
    for (my $count = 0; $count < $#rafsi; $count += 4) {
	my(@r,$s);
	print R "# $count .. ",$count+3,"\n";

	my $ptr = 0;
	$s = "";
	for (my $subcount = 0;
	     $subcount < 4;
	     $subcount++)
	{
	    print R "# ",$count+$subcount,": ";
	    if ($count+$subcount < $#rafsi) {
		my $r = $rafsi[$count+$subcount];
		my $g = rafsi2gismu($r) or die;
		print R "$r -> $g\n";
		$s .= rafsicode($r);
	    } else {
		print R "(none)\n";
		$s .= pack("C3", 0, 0, 0);
	    }
	}
	for (my $subcount = 0;
	     $subcount < 4;
	     $subcount++)
	{
	    if ($count+$subcount < $#rafsi) {
		my $r = $rafsi[$count+$subcount];
		my $g = rafsi2gismu($r) or die;
		my $n = $gismulist{$g};
		die "g: $g, n: $n, r: $r, count: $count, subcount: $subcount"
		  unless defined($n);
		$s .= pack("n",$n);
	    } else {
		$s .= pack("n", 0);
	    }
	}

	printf R "record cat=\@ id=0x00%04X archive\n", $count/4;
	print R "0: ",uc(unpack("H*",$s)),"\n";
	print R "\n";
    }
    close(R);
}
