/* Main code for Gismu Lookup */
/* $Id: gismu.c,v 1.16 1998/11/27 09:18:20 cvs Exp $ */

#include <Pilot.h>
#include "callback.h"

#include "gismu_resources.h"

// ALIGNMENT ALIGNMENT ALIGNMENT!!!! PADDING!
struct GismuRecord {
    Char gismu[5];
    struct {
	unsigned char high;
	unsigned char low;
    } places[5]; // 0 for none, 1 .. 1300ish otherwise
    struct {
	unsigned char high;
	unsigned char low;
    } count;     // 0 for very infrequent
    Char numrafsi;
    Char numrefs;
    Char rafsi[0];
    // from here it gets variable
    // Char rafsi[3][numrafsi];
    // Word refs[numrefs]; // gismu numbers
    // Char description; // null-terminated, two nulls if needed for padding
};

struct GismuCluster {
    unsigned char offsets[4];
    struct GismuRecord firstgismu; // the start of 4 such structs
    // each starting at (char *)cluster + offset[n]
};

struct GismuPlaceCluster {
    Char offsets[4];
    Char firstplace[1]; // the start of 4 null-terminated strings
};

struct RafsiCluster {
    unsigned char rafsi[4][3];
    struct {
	unsigned char high, low;
    } gismu[4];
};

#define gismuAppCreator 'Gsmu'
#define gismuDBType 'GisD'
#define gismuDBName "GismuDB"
#define gismuPlacesDBType 'GisP'
#define gismuPlacesDBName "GismuPlacesDB"
#define rafsiDBType 'Rfsi'
#define rafsiDBName "RafsiDB"

/* globals */
//static const GISMU_PER_CLUSTER = 4;
//static const PLACES_PER_CLUSTER = 4;
#define GISMU_PER_CLUSTER 4
#define PLACES_PER_CLUSTER 4
#define RAFSI_PER_CLUSTER 4
static DmOpenRef GismuDB, GismuPlacesDB, RafsiDB;
static UInt NumClusters;
static UInt NumRafsiClusters;
static char *gismu_c_revision = "$Revision: 1.16 $";
static char *gismu_c_date = "$Date: 1998/11/27 09:18:20 $";
static char GismuListMode = 0;

/* from PalmPower magazine
   <http://www.palmpower.com/issuesprint/issue199807/utility.html>
*/
static void *
GetObjectPtr(Word objectID)
{
    FormPtr pFrm;
    pFrm = FrmGetActiveForm(); // get pointer to the current form
    return(FrmGetObjectPtr(pFrm, FrmGetObjectIndex(pFrm, objectID)));
}

/* return -1 if one comes before two, 0 if they are the same, 1 if one > two */
static signed int
compare_gismu(char one[5], char two[5])
{
    if (one[0] < two[0])
	return (-1);
    if (one[0] > two[0])
	return (+1);
    if (one[1] < two[1])
	return (-1);
    if (one[1] > two[1])
	return (+1);
    if (one[2] < two[2])
	return (-1);
    if (one[2] > two[2])
	return (+1);
    if (one[3] < two[3])
	return (-1);
    if (one[3] > two[3])
	return (+1);
    if (one[4] < two[4])
	return (-1);
    if (one[4] > two[4])
	return (+1);
    // found it
    return (0);
}

static int
lookup_gismu_subindex(int cluster, char target[5])
{
    // search within a cluster for a given gismu. For now the clusters are
    // fixed to have 4 gismu. Upon entering, we know that gismu[0] was tested
    // and the target is later than it. If the target is later than gismu[3],
    // the target doesn't exist.
    // test gismu[2]. If it matches, return 2. If the target is later than [2],
    // test gismu[3] and return 3 or return none. If the target was earlier
    // than [2], test [1] and return 1 or return none.
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *g;
    signed int dir, found;

    r = DmGetRecord(GismuDB, cluster);
    c = (struct GismuCluster *)MemHandleLock(r);
    g = (struct GismuRecord *) ((char *)c + c->offsets[2]);
    dir = compare_gismu(target, &(g->gismu[0]));
    if (dir == 0)
	found = 2;
    else if (dir < 0) {
	// target is earlier than 2: either 1 or nothing
	g = (struct GismuRecord *) ((char *)c + c->offsets[1]);
	if (compare_gismu(target, &(g->gismu[0])))
	    found = -1;
	else
	    found = 1;
    } else {
	// target is later than 2: 3 or nothing
	g = (struct GismuRecord *) ((char *)c + c->offsets[3]);
	if (compare_gismu(target, &(g->gismu[0])))
	    found = -1;
	else
	    found = 3;
    }
    MemHandleUnlock(r);
    DmReleaseRecord(GismuDB, cluster, 0);
    return found;
}

static signed int
lookup_gismu(char target[5])
{
    // do a binary search for a particular gismu. Needs to return a record
    // number and an entry within that record. We could save an extra record
    // open/close by returning the opened record but then we would have to
    // return too many things. quick. dirty. Return -1 for "no such gismu", 0
    // .. max if found. The gismu are guaranteed to be in alphabetical order.
    // There are NumGismu of them.
    int here, top, bottom, dir;
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *g;

    here = NumClusters / 2;
    top = NumClusters;
    bottom = -1;
    while(top > bottom+1) {
	/* check gismu[here] */
	r = DmGetRecord(GismuDB, here);
	c = (struct GismuCluster *)MemHandleLock(r);
	g = (struct GismuRecord *)(&(c->firstgismu));
	// the following could be replaced by
	//  dir = StrNCompare(&(target[0]),&(g->gismu[0]), 5)
	// if we wanted to depend upon PalmOS 2.0 or greater
	dir = compare_gismu(&(target[0]), &(g->gismu[0]));
	MemHandleUnlock(r);
	DmReleaseRecord(GismuDB, here, 0);
	if (dir == +1) {
	    bottom = here;
	    here = (here+top)/2;
	} else if (dir == -1) {
	    top = here;
	    here = (bottom+here)/2;
	} else {
	    // found it
	    return 4*here;
	}
    }
    if (top > 0) {
	// need to check the cluster
	signed int subindex = lookup_gismu_subindex(here, target);
	if (subindex != -1)
	    return 4*here + subindex;
	else
	    return -1;
    }
    // didn't find it.
    return -1;
}

/* return -1 if one comes before two, 0 if they are the same, 1 if one > two */
static signed int
compare_rafsi(unsigned char one[3], unsigned char two[3])
{
    if (one[0] < two[0])
	return (-1);
    if (one[0] > two[0])
	return (+1);
    if (one[1] < two[1])
	return (-1);
    if (one[1] > two[1])
	return (+1);
    if (one[2] < two[2])
	return (-1);
    if (one[2] > two[2])
	return (+1);
    // found it
    return (0);
}

static int
lookup_rafsi_subindex(int cluster, unsigned char target[3])
{
    // search within a cluster for a given rafsi. For now the clusters are
    // fixed to have 4 gismu. Upon entering, we know that rafsi[0] was tested
    // and the target is later than it. If the target is later than rafsi[3],
    // the target doesn't exist.
    // test rafsi[2]. If it matches, return 2. If the target is later than [2],
    // test rafsi[3] and return 3 or return none. If the target was earlier
    // than [2], test [1] and return 1 or return none.
    VoidHand r;
    struct RafsiCluster *c;
    signed int dir, found;

    r = DmGetRecord(RafsiDB, cluster);
    c = (struct RafsiCluster *)MemHandleLock(r);
    dir = compare_rafsi(target, c->rafsi[2]);
    if (dir == 0)
	found = 2;
    else if (dir < 0) {
	// target is earlier than 2: either 1 or nothing
	if (compare_rafsi(target, c->rafsi[1]))
	    found = -1;
	else
	    found = 1;
    } else {
	// target is later than 2: 3 or nothing
	if (compare_rafsi(target, c->rafsi[3]))
	    found = -1;
	else
	    found = 3;
    }
    MemHandleUnlock(r);
    DmReleaseRecord(RafsiDB, cluster, 0);
    return found;
}

static signed int
lookup_rafsi(unsigned char target[3])
{
    // do a binary search for a particular rafsi.
    int here, top, bottom, dir;
    VoidHand r;
    struct RafsiCluster *c;

    here = NumRafsiClusters / 2;
    top = NumRafsiClusters;
    bottom = -1;
    while(top > bottom+1) {
	/* check rafsi[here] */
	r = DmGetRecord(RafsiDB, here);
	c = (struct RafsiCluster *)MemHandleLock(r);
	// the following could be replaced by
	//  dir = StrNCompare(&(target[0]),&(c->rafsi[0]), 3)
	// if we wanted to depend upon PalmOS 2.0 or greater
	dir = compare_rafsi(target, c->rafsi[0]);
	MemHandleUnlock(r);
	DmReleaseRecord(RafsiDB, here, 0);
	if (dir == +1) {
	    bottom = here;
	    here = (here+top)/2;
	} else if (dir == -1) {
	    top = here;
	    here = (bottom+here)/2;
	} else {
	    // found it
	    return 4*here;
	}
    }
    if (top > 0) {
	// need to check the cluster
	signed int subindex = lookup_rafsi_subindex(here, target);
	if (subindex != -1)
	    return 4*here + subindex;
	else
	    return -1;
    }
    // didn't find it.
    return -1;
}

static int
rafsi_to_gismu(int rafsi_index)
{
    VoidHand r;
    struct RafsiCluster *c;
    int gismu_index;

    r = DmGetRecord(RafsiDB, rafsi_index / RAFSI_PER_CLUSTER);
    c = (struct RafsiCluster *)MemHandleLock(r);
    gismu_index = (c->gismu[rafsi_index % RAFSI_PER_CLUSTER].high << 8) |
	c->gismu[rafsi_index % RAFSI_PER_CLUSTER].low;
    MemHandleUnlock(r);
    DmReleaseRecord(RafsiDB, rafsi_index / RAFSI_PER_CLUSTER, 0);
    return gismu_index;
}

static void
display_one_place(struct GismuRecord *p, int which, char Xn, int x, int y)
{
    VoidHand r;
    struct GismuPlaceCluster *place;
    Char *p1, *p2;
    int i, j;
    Word num;

    num = (p->places[which].high << 8) + p->places[which].low;

    if (num) {
	r = DmGetRecord(GismuPlacesDB, num / PLACES_PER_CLUSTER);
	if (r) {
	    int offset;
	    place = (struct GismuPlaceCluster *)MemHandleLock(r);
	    offset = place->offsets[num % PLACES_PER_CLUSTER];
	    p1 = &place->firstplace[offset];
	    // need to turn % into x1 (etc), without sprintf
	    p2 = MemPtrNew(StrLen(p1) + 10);
	    if (p2) {
		for(i=0,j=0; i <= StrLen(p1); i++,j++)
		    if (p1[i] == '%') {
			p2[j++] = 'x';
			p2[j] = Xn;
		    } else
			p2[j] = p1[i];
		WinDrawChars(p2, StrLen(p2), x, y);
		MemPtrFree(p2);
	    } else {
		// out of memory
		WinDrawChars(p1, StrLen(p1), x, y);
	    }
	    MemHandleUnlock(r);
	    DmReleaseRecord(GismuPlacesDB, num / PLACES_PER_CLUSTER, 0);
	}
    }
}

//static const int DEF_X = 40;
//static const int DEF_Y = 20;
#define DEF_X 40
#define DEF_Y 29
#define DEF_SIZE_Y 69

// to test limits: use 'karbi': it has x1..x5, and x5 is wide (needs 120 pxls)
// the rectangle (x 40..160  y 20..90) is for gismu lookups
// index == -1 for "not found"
// draw each place structure element on a separate line: 20,+12,+24,+35,+46,+57
// the second element is the english equivalent: make it stand out
// somehow: bold? more spacing?

static void
display_gismu(int index)
{
    // display gismu[index]
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *p;
    RectangleType rect;
    Char *desc;

    // erase everything, draw boxes
    RctSetRectangle(&rect, DEF_X, 17, 119, 12);
    WinEraseRectangle(&rect, 0);
    WinDrawRectangleFrame(simpleFrame, &rect);

    RctSetRectangle(&rect, DEF_X, DEF_Y, 119, DEF_SIZE_Y);
    WinEraseRectangle(&rect, 0);
    WinDrawRectangleFrame(simpleFrame, &rect);

    if (index == -1) {
	// write "not found" in definition space
	WinSetClip(&rect);
	WinDrawChars("not found", 9, DEF_X+10, DEF_Y+12);
	WinResetClip();
	return;
    } 

    r = DmGetRecord(GismuDB, index / GISMU_PER_CLUSTER);
    if (!r) {
	// write "bad gismu index" in definition space
	WinSetClip(&rect);
	WinDrawChars("bad gismu index", 15, DEF_X+10, DEF_Y+12);
	WinResetClip();
	return;
    }	
    c = (struct GismuCluster *)MemHandleLock(r);
    p = (struct GismuRecord *) ((char *)c + 
				c->offsets[index % GISMU_PER_CLUSTER]);

    RctSetRectangle(&rect, DEF_X, 17, 119, 12);
    WinSetClip(&rect);

    {
	FontID oldfont = FntGetFont();
	FntSetFont(boldFont);
	// width test: 'mamta' is probably the widest gismu
	WinDrawChars(p->gismu, 5, DEF_X+1, 17);
	FntSetFont(oldfont);
    }

    /* build the rafsi string */
    if (p->numrafsi) {
	int i = 0;
	int j = 0;
	int rafsinum;
	Char *rafsibuf = MemPtrNew(17);
	if (rafsibuf) {
	    for (rafsinum = 0; rafsinum < p->numrafsi; rafsinum++) {
		rafsibuf[i++] = p->rafsi[j++];
		rafsibuf[i++] = p->rafsi[j++];
		if (p->rafsi[j] & 0x80)
		    // high bit set means an apostrophe must be inserted
		    rafsibuf[i++] = '\'';
		rafsibuf[i++] = p->rafsi[j++] & 0x7f;
		rafsibuf[i++] = ' '; rafsibuf[i++] = ' ';
	    }
	    rafsibuf[i] = 0;
	    WinDrawChars(rafsibuf, StrLen(rafsibuf), 
			 DEF_X+40, 17);
	}
	MemPtrFree(rafsibuf);
    }
    WinResetClip();

    RctSetRectangle(&rect, DEF_X, DEF_Y, 119, DEF_SIZE_Y);
    WinSetClip(&rect);

    // x1
    display_one_place(p, 0, '1', DEF_X+1, DEF_Y);

    // description: bold is cool, but too wide for many gismu
    desc = (char *)(&p->rafsi) + 3*p->numrafsi + 2*p->numrefs;
    WinDrawChars(desc, StrLen(desc), DEF_X+5, DEF_Y+12);

    // x2
    display_one_place(p, 1, '2', DEF_X+1, DEF_Y+24);

    // x3
    display_one_place(p, 2, '3', DEF_X+1, DEF_Y+35);

    // x4
    display_one_place(p, 3, '4', DEF_X+1, DEF_Y+46);

    // x5
    display_one_place(p, 4, '5', DEF_X+1, DEF_Y+57);

    MemHandleUnlock(r);
    DmReleaseRecord(GismuDB, index / GISMU_PER_CLUSTER, 0);
    WinResetClip();
}

static void
GismuListSelectItem(int index)
{
    ListPtr list;

    if (index < 0)
	index = -index; // approximate position
    list = GetObjectPtr(GismuList);
    LstSetSelection(list, index);
}


static void
GismuListDrawItem(UInt itemNum,
		  RectanglePtr bounds,
		  CharPtr *itemsText)
{
    VoidHand r;
    struct GismuCluster *c;
    struct GismuRecord *p;
    CALLBACK_PROLOGUE;

    r = DmGetRecord(GismuDB, itemNum / GISMU_PER_CLUSTER);
    c = (struct GismuCluster *)MemHandleLock(r);
    p = (struct GismuRecord *) ((char *)c + 
				c->offsets[itemNum % GISMU_PER_CLUSTER]);
    // assume everything will fit.. will this cause problems if it fails?
    WinDrawChars(&(p->gismu[0]), 5, bounds->topLeft.x, bounds->topLeft.y);
    MemHandleUnlock(r);
    DmReleaseRecord(GismuDB, itemNum / GISMU_PER_CLUSTER, 0);
    //WinDrawChars("test",4,bounds->topLeft.x,bounds->topLeft.y);

    CALLBACK_EPILOGUE;
}

static void
RafsiListDrawItem(UInt itemNum,
		  RectanglePtr bounds,
		  CharPtr *itemsText)
{
    VoidHand r;
    struct RafsiCluster *c;
    Char rafsi[4];
    int len;
    CALLBACK_PROLOGUE;

    r = DmGetRecord(RafsiDB, itemNum / RAFSI_PER_CLUSTER);
    c = (struct RafsiCluster *)MemHandleLock(r);
    rafsi[0] = c->rafsi[itemNum % RAFSI_PER_CLUSTER][0];
    rafsi[1] = c->rafsi[itemNum % RAFSI_PER_CLUSTER][1];
    rafsi[2] = c->rafsi[itemNum % RAFSI_PER_CLUSTER][2];
    if (rafsi[2] & 0x80) {
	rafsi[3] = rafsi[2] & 0x7f;
	rafsi[2] = '\'';
	len = 4;
    } else {
	len = 3;
    }

    // assume everything will fit.. will this cause problems if it fails?
    WinDrawChars(rafsi, len, bounds->topLeft.x, bounds->topLeft.y);
    MemHandleUnlock(r);
    DmReleaseRecord(RafsiDB, itemNum / RAFSI_PER_CLUSTER, 0);

    CALLBACK_EPILOGUE;
}

static void
set_list_mode(char mode)
{
    ListPtr list;

    list = GetObjectPtr(GismuList);
    GismuListMode = mode;
    if (mode) {
	// rafsi mode
	LstSetListChoices(list, NULL, NumRafsiClusters * RAFSI_PER_CLUSTER);
	LstSetDrawFunction(list, RafsiListDrawItem);
	LstDrawList(list);
    } else {
	LstSetListChoices(list, NULL, NumClusters * GISMU_PER_CLUSTER); 
	/* there are 1300ish gismu, each record is a cluster of 4 */
	LstSetDrawFunction(list, GismuListDrawItem);
	LstDrawList(list);
    }
}

static void
DoAboutBox(void)
{
    FormPtr frm;

    frm = FrmInitForm(GismuAboutForm);
    /* substitute in version numbers whereever necessary */
    
    FrmDoDialog(frm);
    FrmDeleteForm(frm);
}

static void
MainFormInit(void)
{
    FormPtr frm;
    ControlPtr button;

    frm = FrmGetActiveForm();
    /* set up the list */
    set_list_mode(0);
    button = GetObjectPtr(GismuListGismuButton);
    CtlSetValue(button, 1);
    FrmDrawForm(frm);
}

static Boolean 
MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (e->eType) {
      case frmOpenEvent:
	MainFormInit();
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	  case MainOptionsAbout:
	    DoAboutBox();
	}

    	handled = true;
	break;

      case lstSelectEvent: // an entry in the gismu list was selected
	  {
	      ListPtr list;
	      int index;

	      list = GetObjectPtr(GismuList);
	      index = LstGetSelection(list);
	      if (GismuListMode) {
		  // rafsi mode
		  index = rafsi_to_gismu(index);
	      }
	      display_gismu(index);
	      
	      handled = true;
	  }

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	  case GismuListGismuButton :
	    set_list_mode(0);
	    handled = true;
	    break;
	  case GismuListRafsiButton :
	    set_list_mode(1);
	    handled = true;
	    break;
	  case GismuLookupButton : 
	      {
		  Handle t;
		  FieldPtr entry_ptr;
		  unsigned char *c;
		  unsigned char entry[5] = {0,0,0,0,0};
		  int i, gismu_index;

		  // first, read the text out of the entry field
		  entry_ptr = GetObjectPtr(GismuEntry);
		  t = FldGetTextHandle(entry_ptr);
		  c = MemHandleLock((VoidHand)t);
		  //WinDrawChars(c,StrLen(c), 50, 100);
		  for(i=0; i < 5 && i < StrLen(c); i++)
		      entry[i] = c[i];
		  MemHandleUnlock((VoidHand)t);

		  if (GismuListMode) {
		      // rafsi mode
		      if (StrLen(c) == 4) {
			  entry[2] = entry[3] | 0x80;
		      }
		      i = lookup_rafsi(entry);
		      GismuListSelectItem(i);
		      if (i < 0)
			  gismu_index = -1;
		      else
			  gismu_index = rafsi_to_gismu(i);
		      display_gismu(gismu_index);
		  } else {
		      // gismu mode
		      gismu_index = lookup_gismu(entry);
		      display_gismu(gismu_index);
		      GismuListSelectItem(gismu_index);
		  }
		  handled = true;
	      }
	    break;
	}
	break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean 
ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word 
StartApplication(void)
{
    GismuDB = DmOpenDatabaseByTypeCreator(gismuDBType, 
					  gismuAppCreator, 
					  dmModeReadOnly);
    if (!GismuDB) {
	// signal an error: this database must be constructed externally
	FrmAlert(MissingGismuDBAlert);
	// PalmOS 1.0 will relaunch this app, but hitting the "Applications"
	// button will let them get away.
	return -1;
    }
    GismuPlacesDB = DmOpenDatabaseByTypeCreator(gismuPlacesDBType, 
						gismuAppCreator,
						dmModeReadOnly);
    if (!GismuPlacesDB) {
	FrmAlert(MissingGismuPlacesDBAlert);
	if (GismuDB)
	    DmCloseDatabase(GismuDB);
	return -1;
    }
    NumClusters = DmNumRecords(GismuDB);
    RafsiDB = DmOpenDatabaseByTypeCreator(rafsiDBType, 
					  gismuAppCreator,
					  dmModeReadOnly);
    if (!RafsiDB) {
	FrmAlert(MissingRafsiDBAlert);
	if (GismuDB)
	    DmCloseDatabase(GismuDB);
	if (GismuPlacesDB)
	    DmCloseDatabase(GismuPlacesDB);
	return -1;
    }
    NumRafsiClusters = DmNumRecords(RafsiDB);
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void 
StopApplication(void)
{
    DmCloseDatabase(GismuDB);
    DmCloseDatabase(GismuPlacesDB);
    DmCloseDatabase(RafsiDB);
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void 
EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord 
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}

/* todo:
   catch a "return" in the gismu entry field and treat it like a button press
    (so I don't have to jump between the graffiti area and the button on top)
    probably need to watch for keyDownEvent
   selecting a gismu in the list sets the field text to that gismu.
   display more info about the gismu. Need to add more info to the database
    first. How to do it space-efficiently?
   change the GismuPlacesDB to use the length of the record instead of having
    them all null-terminated (would save 1.3k off the 58.5k database)
   hitting up/down buttons should scroll gismu list
   display xrefs
   make xrefs into hyperlinks
   deal with partial clusters, particularly in RafsiDB (which has them)
   maybe put search list on the right (so a right-handed person like me doesn't
    obscure the definition to pick new gismu): make it configurable?
   sidju: helps
   remember the last-looked-up gismu and last rafsi. Return to that one when
    changing modes. Or, when a gismu is displayed, set the current gismu to it,
    and if it has rafsi (and we're in gismu mode), set the current rafsi to 
    the first rafsi. Remember the current mode and word across invocations.
   when writing a word that isn't quite matched, jump to somewhere nearby in
    the wordlist

   flashcard stuff. same app? menu-bar mode selector?

 */
